/**
 * \file motor.cpp
 * \brief 简要介绍
 * \details 详细说明
 * \author lenotary
 * \version 1.0
 * \date 2021/9/13
 * \pre 先决条件。（有的话才添加。）
 * \bug 存在的bug。（有的话才添加，注明“还未发现”也可以。）
 * \warning 警告。（有的话才添加。）
 * \copyright 版权声明。（通常只写遵循什么协议，版权详细内容应放在LICENSE文件中，并且应该与LICENSE文件的内容统一，不能自相矛盾。）
 * \since 一些历史情况记录。（有的话才添加。）
 */

//
// Created by lenotary on 2021/9/13.
//

#include "motor.h"

Motor::Motor(ros::NodeHandle &nh, ros::NodeHandle &privateNh) {
    m_iPid_ = new Pid(150.0, 2.0, 0.0, 2000, 2000);
    m_iPubPose_ = nh.advertise<std_msgs::Float64>("motor_pose", 1);
    m_iPubPoseImu_ = nh.advertise<sensor_msgs::Imu>("motor_pose_imu", 1);
    m_iPubSpeed_ = nh.advertise<std_msgs::Float64>("motor_speed", 1);
    m_iPubCurrent_ = nh.advertise<std_msgs::Float64>("motor_current", 1);
    m_iSubStart_ = nh.subscribe("start_motor", 1, &Motor::setSpeedCallback, this);
    m_iSubPose_ = nh.subscribe("motor_set_pose", 1, &Motor::setPoseCallback, this);
    privateNh.getParam("speed", m_dParamSpeed);
    struct sockaddr_can iAddress{};
    struct ifreq ifr{};
    struct can_filter rFilter{};
    m_nS_ = socket(PF_CAN, SOCK_RAW, CAN_RAW);//创建套接字
    strcpy(ifr.ifr_name, "can0");
    ioctl(m_nS_, SIOCGIFINDEX, &ifr); //指定 can0 设备
    iAddress.can_family = AF_CAN;
    iAddress.can_ifindex = ifr.ifr_ifindex;
    bind(m_nS_, (struct sockaddr *) &iAddress, sizeof(iAddress));//将套接字与 can0 绑定
    setsockopt(m_nS_, SOL_CAN_RAW, CAN_RAW_FILTER, &rFilter, sizeof(rFilter));
    std::thread t(&Motor::receiveProcess, this);
    t.detach();
}


void Motor::receiveProcess() {
    uint64_t count{0};
    while (ros::ok()) {
        m_iSpeedMutex_.lock();
        m_nBytes_ = read(m_nS_, &m_iGetFrame_, sizeof(m_iGetFrame_));
        if (m_nBytes_ > 0) {
            m_iMotorInfo_.m_nMotorId = m_iGetFrame_.can_id;
            m_iMotorInfo_.GetPose((m_iGetFrame_.data[0] << 8) + m_iGetFrame_.data[1]);
            m_iMotorInfo_.GetSpeed((m_iGetFrame_.data[2] << 8) + m_iGetFrame_.data[3]);
            m_iMotorInfo_.GetTorqueCurrent((m_iGetFrame_.data[4] << 8) + m_iGetFrame_.data[5]);
            m_iMotorInfo_.GetTemperature(m_iGetFrame_.data[6]);
        }
        static double prevAngle = m_iMotorInfo_.m_dPoseAngle;
        static double prevTime = ros::Time::now().toSec();
        if (std::abs(prevAngle - m_iMotorInfo_.m_dPoseAngle) > M_PI and
            std::abs(prevTime - ros::Time::now().toSec()) > 1.0) {
            ROS_WARN("info %f %f", prevAngle, m_iMotorInfo_.m_dPoseAngle);
            m_dSetSpeed = -m_dSetSpeed;
            prevTime = ros::Time::now().toSec();
        }
        prevAngle = m_iMotorInfo_.m_dPoseAngle;

        if( std::abs(m_dSetPose_) > M_PI / 2.0 ){
            if(m_dSetPose_ < 0){
                m_dSetPose_ = 2 * M_PI + m_dSetPose_;
            }
            m_iMotorInfo_.m_nSetVoltage = m_iPid_->PidCalc(m_iMotorInfo_.poseRadToInt(m_dSetPose_),
                                                           m_iMotorInfo_.m_nPoseInt2);
//            ROS_INFO("info ss %d %d",m_iMotorInfo_.poseRadToInt(m_dSetPose_),m_iMotorInfo_.m_nPoseInt2);
        }else{
            m_iMotorInfo_.m_nSetVoltage = m_iPid_->PidCalc(m_iMotorInfo_.poseRadToInt(m_dSetPose_),
                                                           m_iMotorInfo_.m_nPoseInt);
//            ROS_INFO("info ss %d %d",m_iMotorInfo_.poseRadToInt(m_dSetPose_),m_iMotorInfo_.m_nPoseInt);
        }

        sendMsg(1, static_cast<short >( m_iMotorInfo_.m_nSetVoltage));
        std_msgs::Float64 msg;
        msg.data = m_iMotorInfo_.m_dPoseAngle;
        m_iPubPose_.publish(msg);
        msg.data = m_iMotorInfo_.m_dSpeedRad;
        m_iPubSpeed_.publish(msg);
        msg.data = m_iMotorInfo_.m_dTorqueCurrent;
        m_iPubCurrent_.publish(msg);

        sensor_msgs::Imu imuMsg;
        imuMsg.header.seq = count++;
        imuMsg.header.stamp = ros::Time::now();
        imuMsg.header.frame_id = "motor";
        auto q = tf::createQuaternionFromYaw(m_iMotorInfo_.m_dPoseAngle);
        imuMsg.orientation.x = q.x();
        imuMsg.orientation.y = q.y();
        imuMsg.orientation.z = q.z();
        imuMsg.orientation.w = q.w();
        m_iPubPoseImu_.publish(imuMsg);
        m_iSpeedMutex_.unlock();
    }
}

void Motor::sendMsg(uint32_t id, short data) {
    struct can_frame sendFrame{};
    sendFrame.can_id = 0x1ff;
    sendFrame.can_dlc = 8;
    sendFrame.data[0] = (data >> 8) & 0xff;
    sendFrame.data[1] = (data) & 0xff;

    m_nBytes_ = write(m_nS_, &sendFrame, sizeof(sendFrame));
    if (m_nBytes_ != sizeof(sendFrame)) {
        ROS_WARN("send error");
    }
}

void Motor::setSpeedCallback(const std_msgs::Bool::ConstPtr &msg) {
    m_iSpeedMutex_.lock();
    m_bIsRun_ = msg->data;
//    if(msg->data == true){
//        m_dSetSpeed = m_dParamSpeed;
//    }else{
//        m_dSetSpeed = 0;
//    }
    m_iSpeedMutex_.unlock();
}
void Motor::setPoseCallback(const std_msgs::Float64::ConstPtr &msg) {
    m_iSpeedMutex_.lock();
    ROS_INFO("[motor] get pose %f",msg->data);
    m_dSetPose_ = msg->data;
    if(m_dSetPose_ > M_PI){
        m_dSetPose_ = M_PI;
    }else if(m_dSetPose_ < -M_PI){
        m_dSetPose_ = -M_PI;
    }
    m_iSpeedMutex_.unlock();
}
