/**
 * \file motor.h
 * \brief 简要介绍
 * \details 详细说明
 * \author lenotary
 * \version 1.0
 * \date 2021/9/13
 * \pre 先决条件。（有的话才添加。）
 * \bug 存在的bug。（有的话才添加，注明“还未发现”也可以。）
 * \warning 警告。（有的话才添加。）
 * \copyright 版权声明。（通常只写遵循什么协议，版权详细内容应放在LICENSE文件中，并且应该与LICENSE文件的内容统一，不能自相矛盾。）
 * \since 一些历史情况记录。（有的话才添加。）
 */


//
// Created by lenotary on 2021/9/13.
//

#ifndef SRC_MOTOR_H
#define SRC_MOTOR_H

#include <ros/ros.h>
#include <ros/node_handle.h>
#include <tf2/transform_datatypes.h>
#include <tf/tf.h>
#include <linux/can.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/can/raw.h>
#include <thread>
#include <cmath>
#include <std_msgs/Float64.h>
#include <std_msgs/Bool.h>
#include <sensor_msgs/Imu.h>
#include <mutex>
#include "pid.h"


class MotorInfo {
public:
    uint32_t m_nMotorId;
    double m_nSetVoltage;
    short m_nSetSpeedInt;
    short m_nSpeedInt;
    int m_nPoseInt;
    int m_nPoseInt2;
    short m_nTorqueCurrentInt;
    short m_nTemperatureInt;
    double m_dSpeedRad;
    double m_dPoseAngle;
    double m_dTorqueCurrent;

    short speedRadToInt(double input) {
        return static_cast<short>(input * 60.0 / M_PI);
    }
    short poseRadToInt(double input) {
        return static_cast<short>(input * 8192.0 / 2 / M_PI);
    }

    void GetSpeed(short input) {
        m_nSpeedInt = input;
        m_dSpeedRad = m_nSpeedInt / 60.0 * M_PI;
    }

    void GetPose(short input) {
        m_nPoseInt = input;
        m_nPoseInt2 = input;
        if(input > 8192 / 2){
            m_nPoseInt = input - 8192;
        }

        m_dPoseAngle = input / 8192.0 * 2 * M_PI;
        if (m_dPoseAngle > M_PI) {
            m_dPoseAngle = m_dPoseAngle - 2 * M_PI;
        }
    }

    void GetTorqueCurrent(short input) {
        m_nTorqueCurrentInt = input;
        m_dTorqueCurrent = input / 1000.0;
    }

    void GetTemperature(short input) {
        m_nTemperatureInt = input;
    }

    void SetSpeed(double input) {
        m_nSetSpeedInt = static_cast<short >(input * 60.0 / M_PI);
    }

};

class Motor {
public:
    Motor(ros::NodeHandle &nh, ros::NodeHandle &privateNh);

private:
    int m_nS_, m_nBytes_;
    ros::Publisher m_iPubPoseImu_;
    ros::Publisher m_iPubPose_;
    ros::Publisher m_iPubSpeed_;
    ros::Publisher m_iPubCurrent_;
    ros::Subscriber m_iSubStart_;
    ros::Subscriber m_iSubPose_;
    Pid *m_iPid_;
    MotorInfo m_iMotorInfo_;
    std::mutex m_iSpeedMutex_;
    struct canfd_frame m_iGetFrame_{};
    double m_dSetPose_{0.0};
    bool m_bIsRun_;
    double m_dSetSpeed{0};
    double m_dParamSpeed{0};
    void receiveProcess();
    void sendMsg(uint32_t id, short data);
    void setSpeedCallback(const std_msgs::Bool::ConstPtr &msg);
    void setPoseCallback(const std_msgs::Float64::ConstPtr &msg);
};


#endif //SRC_MOTOR_H
